/***************************************************************************
 *   Copyright (C) 2008 by Travis   *
 *   mopar@linux-laptop   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <vector>
#include <list>
#include <fstream>

using namespace std;

#include "markovchains.h"

MarkovChains::MarkovChains(){
	srand((unsigned)time(0));
}

void MarkovChains::trim(string &str)const
{
	string::size_type pos = str.find_last_not_of(' ');
	if(pos != string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(' ');
		if(pos != string::npos) 
			str.erase(0, pos);
	}
	else 
		str.erase(str.begin(), str.end());
}

//adds every sentance in a text file to the brain one at a time
void MarkovChains::addFile(string fileName )
{
	ifstream file( fileName.c_str() );
	if ( !file ){
		cout << "Couldn't open input file: " << fileName << endl;
		return;
	}

	string buffer;
	char ch;
	while (file.get(ch)) {
		if(ch == '\r' || ch == '\n')
			ch = ' ';
		buffer.push_back(ch);
		if (endChars.find(ch) != string::npos) {
			add(buffer);
			buffer.clear();
		}
	}
	add(buffer);
	file.clear();
	file.close();
}

//adds a single sentance to the brain
void MarkovChains::add(string sentence){
	trim(sentence);
//	cout << "adding: " << sentence << endl;
	vector<string> parts;
	const char* chars = sentence.c_str();
	int i = 0;
	bool punctuation = false;
	string buffer;
	//split the sentance up into parts
	while (i < sentence.length()) {
		char ch = chars[i];
		if ((validChars.find(ch) != string::npos) == punctuation) {
			punctuation = !punctuation;
			if (buffer.length() > 0) {
				parts.push_back(buffer);
			}
			buffer.clear();
			continue;
		}
		buffer.push_back(ch);
		i++;
	}
	//add last buffer
	if (buffer.length() > 0)
		parts.push_back(buffer);
	
	//didnt learn anything
	if (parts.size() < 4) 
		return;
	
	//may have learned something depending on whether or not we know it already...
	for (i = 0; i < parts.size() - 3; i++) {
		Quad *quad = new Quad(parts.at(i), parts.at(i + 1), parts.at(i + 2), parts.at(i + 3));
		//if we already have this quad, get the old one and delete what we just made, we may have to update the old ones canEnd and/or canStart
		if (quads.count(*quad) > 0) {
			Quad *temp = quads.find(*quad)->second;
			delete quad;
			quad = temp;
		}
		//otherwise its new, save it
		else {
			quads.insert( pair<Quad, Quad*>(*quad, quad) );
		}

		if (i == 0)
			quad->canStart = true;
			
		if (i == parts.size() - 4)
			quad->canEnd = true;

		//store it in the words map so that we can get quads based on user input
		for (int n = 0; n < 4; n++) {
			string token = parts.at(i + n);
			if (words.count(token) == 0)
				words.insert( pair<string, map<Quad, Quad*>* >(token, new map<Quad, Quad*>));
			(words.find(token)->second)->insert( pair<Quad, Quad*>(*quad, quad) );
		}

		//update the previous map
		if (i > 0) {
			string previousToken = parts.at(i - 1);
			if (previous.count(*quad) == 0)
				previous.insert( pair<Quad, set<string>* >(*quad, new set<string>));
			(previous.find(*quad)->second)->insert(previousToken);
		}

		//update the next map
		if (i < parts.size() - 4) {
			string nextToken = parts.at(i + 4);
			if (next.count(*quad) == 0)
				next.insert( pair<Quad, set<string>* >(*quad, new set<string>));
			(next.find(*quad)->second)->insert(nextToken);
		}
	}
}

//get a random sentance
string MarkovChains::getSentence(){
	return getSentence("");
}

//get a sentance with a given word in it
string MarkovChains::getSentence(string word){
	map<Quad, Quad*> *quadstouse;
	//if the word can be found, get all quads that word is in
	if (words.count(word) > 0) {
		quadstouse = words.find(word)->second;
	}
	//otherwise get all quads we have
	else {
		quadstouse = &quads;
	}
	//load the selected quads into an array for easy access
	vector<Quad*> quadArr;
	for(map<Quad, Quad*>::iterator iter = quadstouse->begin(); iter != quadstouse->end(); iter++ )
		quadArr.push_back(iter->second);//be sure to get the second, the first isnt updated remember?

	//if there are no quads, return an empty string
	if (quadArr.size() == 0)
		return "";

	//pick a random quad, save the reference for later
	Quad *randomQuad = quadArr.at(rand() % quadArr.size());
	Quad *quad = randomQuad;
	
	//load all the words from the quad into the sentance parts
	list<string> parts;
	for (int i = 0; i < 4; i++)
		parts.push_back(quad->tokens[i]);

	//build the end of the sentance
	while (quad->canEnd == false) {
		//get all possible words that can come after this quad
		set<string> *stringSet = next.find(*quad)->second;
		//load them into a vector
		vector<string> nextTokens;
		for(set<string>::iterator  theIterator = stringSet->begin(); theIterator != stringSet->end(); theIterator++ )
			nextTokens.push_back(*theIterator);
		//choose a word at random and put it at the end of the sentance parts
		string nextToken = nextTokens.at(rand() % nextTokens.size());
		parts.push_back(nextToken);
		//set quad to the quad that ends with the above randomly chosen token
		Quad tempQuad(quad->tokens[1], quad->tokens[2], quad->tokens[3], nextToken);
		quad = quads.find(tempQuad)->second;
	}

	//reset the quad pointer to the original random quad
	quad = randomQuad;
	//build the beginning of the sentance
	while (quad->canStart == false) {
		//get all possible words that can come before this quad
		set<string> *stringSet = previous.find(*quad)->second;
		//load them into a vector
		vector<string> previousTokens;
		for(set<string>::iterator theIterator = stringSet->begin(); theIterator != stringSet->end(); theIterator++ )
			previousTokens.push_back(*theIterator);
		//choose a word at random and put it at the beginning of the sentance parts
		string previousToken = previousTokens.at(rand() % previousTokens.size());
		parts.push_front(previousToken);
		//set quad to the quad that begins with the above randomly chosen token
		Quad tempQuad(previousToken, quad->tokens[0], quad->tokens[1], quad->tokens[2]);
		quad = quads.find(tempQuad)->second;
	}
	
	//put the sentance together
	string sentence;
	for(list<string>::iterator theIterator = parts.begin(); theIterator != parts.end(); theIterator++ )
		sentence.append(*theIterator);

	return sentence;
}

int main(int argc, char *argv[])
{
	MarkovChains mark;
	string input;
	string firstWord;

	cout << "commands:\n1. 'exit' exits the program\n2. 'load filename' loads all sentances in filename into the system (test.txt is provided)\n3. Any other input is loaded into the system, and a sentance is output based on the first word typed if possible" << endl;

	while(true){
		cout << "# ";
		getline(cin, input);
		firstWord = input.substr(0, input.find(' '));
		if(firstWord == "exit")
			break;
		else if(firstWord == "load")
			mark.addFile(input.substr(input.find(' ')+1, input.length()));
		else
			mark.add(input);
		cout << mark.getSentence(firstWord) << endl;
	}

	return EXIT_SUCCESS;
}
