/***************************************************************************
 *   Copyright (C) 2008 by Travis   *
 *   mopar@linux-laptop   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <map>
#include <set>

using namespace std;

const int order = 4;

const string validChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
const string endChars = ".!?";

class Quad {
public:
	Quad(string s1, string s2, string s3, string s4){
		tokens[0] = s1;
		tokens[1] = s2;
		tokens[2] = s3;
		tokens[3] = s4;
		canStart = false;
		canEnd = false;
	}
	bool operator<(Quad other)const{
		for(int x = 0; x < order; x++){
			if(tokens[x] < other.tokens[x])
				return true;
			if(other.tokens[x] < tokens[x])
				return false;
		}
		return false;
	}
	bool operator==(Quad other)const{
		return tokens[0] == other.tokens[0] && tokens[1] == other.tokens[1] && tokens[2] == other.tokens[2] && tokens[3] == other.tokens[3];
	}	
	string tokens[order];
	bool canStart;
	bool canEnd;
};

class MarkovChains {
public:
	MarkovChains();
	//adds every sentance in a text file to the brain one at a time
	void addFile(string);
	//adds a single sentance to the brain
	void add(string);
	//get a random sentance
	string getSentence();
	//get a sentance with a given word in it
	string getSentence(string);
private:
	void trim(string&)const;
	map<string, map<Quad, Quad*>* > words;
	map<Quad, Quad*> quads;
	map<Quad, set<string>* > next;
	map<Quad, set<string>* > previous;
};
